!=====================================================================
!
!       m e m b r a n e S p h e r e  1 . 0
!       --------------------------------------------------
!
!      Daniel Peter
!      ETH Zurich - Institute of Geophysics
!      (c) ETH July 2006
!
!      Free for non-commercial academic research ONLY.
!      This program is distributed WITHOUT ANY WARRANTY whatsoever.
!      
!=====================================================================

!-----------------------------------------------------------------------
      subroutine forwardIteration()
!-----------------------------------------------------------------------
! iterates through time and calculates the solutions at each time step
!
! finite-difference iteration: information in carl tape thesis2003, chap 5, (5.7)
      use propagationStartup; use cells; use phaseVelocityMap; use parallel; use displacements
      use phaseBlockData; use griddomain; use adjointVariables; use verbosity
      implicit none
      real(WP):: u_t, u_tplus1, u_tminus1,forcing,initialShape,forcingRef
      real(WP):: D2,time,discreteLaplacian, precalc_discreteLaplacian
      real(WP):: forceTerm2Source,initialShapeTerm,lat,lon,newdisp
      integer:: n,i,k,timestep,vertex,index,ioerror
      external:: initialShapeTerm,discreteLaplacian,precalc_discreteLaplacian,integrand             
      character*4:: timestr
      real(WP):: vector(3),vsource(3),vreceiver(3),rotationMatrix(3,3)
      integer:: forceVertex
      
      !initialize
      displacement(:)    =0.0_WP
      displacement_old(:)=0.0_WP
      newdisplacement(:) =0.0_WP
      
!      ! recalculate the rotation matrix for the right source indexing in the prescribed array
!      if( Adjoint_InversionProgram ) then
!        vsource(:) = vertices(sourceVertex,:)
!        vreceiver(:)= vertices(receiverVertex,:)
!        call getInverseRotationMatrix( vsource,vreceiver,rotationMatrix )
!      endif
      
      
      ! time iteration of displacements
      index=0
      do timestep= firsttimestep, lasttimestep
        ! model time
        time = timestep*dt      
        !debug 
        !if(DEBUG) then
        !  print*,'    time:',rank,time
        !endif        
        
        index=index+1

        ! swap displacement arrays
        displacement_old(:) = displacement(:)
        displacement(:) = newdisplacement(:)
                   
        ! propagate only corresponding vertices
        do n=1, numDomainVertices
          ! choose vertex
          if( PARALLELSEISMO ) then
            vertex=domainVertices(n)
          else
            vertex=n
          endif
          
          ! spherical Laplacian    
          if( PRECALCULATED_CELLS ) then      
            D2 = precalc_discreteLaplacian(vertex) !uses displacement(..) field
          else
            D2 = discreteLaplacian(vertex)
          endif
          ! calculate new displacement in time
          u_t = displacement(vertex)
          u_tminus1 = displacement_old(vertex)

          ! determine force
          if( PRESCRIBEDSOURCE ) then
            ! get the value out of the prescribed force array
!            if( Adjoint_InversionProgram ) then
!              ! determine the index in the prescribed array            
!              vector(:) = vertices(vertex,:)
!              call rotateVector(rotationMatrix,vector,vector)              
!              call getSphericalCoordinates(vector,lat,lon)              
!              call findVertex(lat,lon,forceVertex)
!
!              ! use the force, luke.
!              forcing = forceTermPrescribed(forceVertex,index)                          
!            else
              forcing = forceTermPrescribed(n,index)  ! prescribed array goes on till numDomainVertices, see indexing at initialization          
!            endif            
          else
            forcing = forceTerm2Source(vertex,time,sourceVertex)
          endif

          ! store for output
          !if( vertex .eq. receiverVertex) forcingRef = forcing
          
          ! calculate phase velocity square 
          cphase2 = phaseVelocitySquare(vertex)
                           
          !propagation step                                                     
          u_tplus1=u_t+u_t-u_tminus1+cphase2*dt2*(D2+forcing)
                        
          ! iterated displacements
          newdisplacement(vertex) = u_tplus1
          
          !debug
          if(DEBUG) then
            !check if NaN
            if( u_tplus1 .ne. u_tplus1) then
              !u_tplus1 is not a number
              print*,'u_tplus1:',u_tplus1
              print*,u_t, u_tminus1,cphase2,dt2,D2,forcing
              call stopProgram( 'abort - program phaseshift displacement is NaN     ')
            endif                          
          endif
        enddo

        if( PARALLELSEISMO ) then
          ! synchronize new displacement arrays          
          call syncNewdisplacement()
        endif
                            
        ! fill displacement at receiver station into seismogram record of that receiver station
        !if(DEBUG) print*,'    recording at receiver...',rank,index,time,newdisplacement(receiverVertex)
        ! note: newdisplacement is at time step t+dt        
        if( manyReceivers ) then
          call recordManySeismograms(index,time+dt,referenceRun)
        else
          ! set time in seismogram
          seismogramReceiver(1,index)=time+dt
        
          ! set displacement
          if( Station_Correction ) then
            !interpolate new displacement for original receiver location
            call interpolateLinear(interpolation_distances,interpolation_triangleLengths,&
                                  interpolation_corners,newdisp)
            seismogramReceiver(2,index)=newdisp                                      
          else
            seismogramReceiver(2,index)=newdisplacement(receiverVertex)            
          endif
        endif

        !file output for simulation snapshots
        if(SIMULATIONOUTPUT) then
          if( mod(timestep,10) .eq. 0 .and. time .ge. 0.0 .and. MASTER ) then
            write(timestr,'(i4.4)') timestep
            open(10,file=datadirectory(1:len_trim(datadirectory))//'simulation.'//timestr//'.dat')        
            do n=1, numVertices
              write(10,'(4f18.6)')(vertices(n,k),k=1,3),newdisplacement(n)
            enddo
            close(10) 
          endif          
        endif
        
        ! save displacements at each time step for future adjoint calculation
        !if( Adjoint_Program .and. .not. ADJOINT_ONTHEFLY ) then
        if( Adjoint_Program ) then
          call storeForwardDisplacements(timestep,index)
        endif        
      enddo !timestep      
      end

!-----------------------------------------------------------------------   
      subroutine recordManySeismograms(index,time,refRun)
!-----------------------------------------------------------------------   
! fill receivers seismograms with computed displacements
      use propagationStartup; use displacements
      implicit none
      logical:: refRun
      integer:: index,m,i
      real(WP):: time
      
      ! record seismograms   
      if( manyKernels ) then
        if( refRun ) then
          do m=1,numofKernels
            ! fill in the time seismogram first
            kernelsReceiversSeismogramRef(m,numofReceivers+1,index)=time
            ! fill in the displacement seismograms for each receiver
            do i=1,numofReceivers
              kernelsReceiversSeismogramRef(m,i,index)=newdisplacement(kernelsReceivers(m,i))
            enddo
          enddo
        else
          do m=1,numofKernels
            ! fill in the time seismogram first
            kernelsReceiversSeismogram(m,numofReceivers+1,index)=time
            ! fill in the displacement seismograms for each receiver
            do i=1,numofReceivers
              kernelsReceiversSeismogram(m,i,index)=newdisplacement(kernelsReceivers(m,i))
            enddo              
          enddo
        endif
      else            
        if( refRun ) then
          ! fill in the time seismogram first
          receiversSeismogramRef(numofReceivers+1,index)=time
          ! fill in the displacement seismograms for each receiver
          do i=1,numofReceivers
            receiversSeismogramRef(i,index)=newdisplacement(receivers(i))
          enddo
        else
          ! fill in the time seismogram first
          receiversSeismogram(numofReceivers+1,index)=time
          ! fill in the displacement seismograms for each receiver
          do i=1,numofReceivers
            receiversSeismogram(i,index)=newdisplacement(receivers(i))
          enddo              
        endif
      endif
      end


!-----------------------------------------------------------------------
      subroutine parallelFindnextLocation(move)
!-----------------------------------------------------------------------
! increments delta location and finds new delta vertex
! (remember: different processes are used to compute the same simulation, not different simulations)
!
! input:
!     latDelta/lonDelta     - location of delta
!     vertexDelta             - nearest vertex
!     move                      - iteration flag
!
! return: latDelta,lonDelta,vertexDelta,move for new location
      use loop; use propagationStartup; use deltaSecondLocation
      implicit none
      logical:: move
      
      ! increment latitude
      if( abs(deltaLat-latitudeEnd) .gt. deltaMoveIncrement/2 ) then
        deltaLat=deltaLat+deltaMoveIncrement
      else
        ! increment longitude
        ! if( int(latDelta) .gt. 50) then
        deltaLat= latitudeStart
        deltaLon= deltaLon+deltaMoveIncrement
      endif
      
      ! get nearest vertex for new location
      call findVertex(deltaLat,deltaLon,deltaVertex)

      ! for 2 delta scatterers
      if( SECONDDELTA) call placeSecondDelta(deltaLat,deltaLon,deltaSecondLat, &
                            deltaSecondLon,deltaSecondVertex,deltaSecondDistance)
  
      
      !stop iteration for high longitude
      if( deltaLon .gt. (longitudeEnd + deltaMoveIncrement/2) ) move=.false.
      
      end
      
!-----------------------------------------------------------------------
      subroutine findnextLocation(move)
!-----------------------------------------------------------------------
! increments delta location for each process and finds new delta vertex
! (i.e. each process runs a different simulation)
! input:
!     latDelta/lonDelta     - location of delta
!     vertexDelta             - nearest vertex
!     move                      - iteration flag
!
! return: latDelta,lonDelta,vertexDelta,move for new location
      use loop; use parallel; use propagationStartup; use deltaSecondLocation
      implicit none
      real(WP):: lat,lon,newlon,olddeltaLon
      logical:: move,bestfound
      integer:: olddeltaVertex
            
      ! increment latitude
      if( (deltaLat - (latitudeEnd-nprocesses*deltaMoveIncrement) ) .lt. deltaMoveIncrement/2 ) then
        ! not yet reached the end, so increment
        deltaLat=deltaLat+nprocesses*deltaMoveIncrement
      else
        ! deltaLat is at the latitudeEnd or has already passed it
        if( manyReceivers ) then
          ! moves only down the latitude at one chosen longitude, when end reached, it stops
          move=.false.
        else
          ! reset latitude to begining (take account of alternating which process will start from latitudeStart)
          deltaLat= latitudeStart+( deltaLat-latitudeEnd+(nprocesses-1)*deltaMoveIncrement)
          ! increment longitude
          deltaLon=deltaLon+deltaMoveIncrement

          !stop iteration for high longitude
          if( (deltaLon - longitudeEnd) .gt. deltaMoveIncrement/2 ) move=.false.        
        endif
      endif
            
      ! get nearest vertex for new location
      call findVertex(deltaLat,deltaLon,deltaVertex)
      
      ! avoid too big steps
      if( manyReceivers ) then
        call getSphericalCoord_Lat(deltaVertex,lat,lon)
        newlon=deltaLon
        bestfound=.false.
        if( abs(lat-deltaLat) .le. abs(deltaMoveIncrement/2)) bestfound=.true.

        do while( (.not. bestfound ) .and. (newlon .lt. (deltaLon+360.0))  )
          ! search on the east for a better vertex
          newlon=newlon+deltaMoveIncrement*0.5
          call findVertex(deltaLat,newlon,deltaVertex)
          call getSphericalCoord_Lat(deltaVertex,lat,lon)          
          if( abs(lat-deltaLat) .le. abs(deltaMoveIncrement/2) ) bestfound=.true.
        enddo              
        ! ...we found the best vertex (best matching latitude)
        if( .not. bestfound ) then
          ! take initial one
          call findVertex(deltaLat,deltaLon,deltaVertex)
          call getSphericalCoord_Lat(deltaVertex,lat,lon)          
          ! output
          print*
          print*,'no a very good delta vertex found for position:',deltaLat,deltaLon
          print*,'   found only:',lat,lon
          print*
        endif
        
        ! for 2 delta scatterers
        if( SECONDDELTA) call placeSecondDelta(deltaLat,newlon,deltaSecondLat, &
                              deltaSecondLon,deltaSecondVertex,deltaSecondDistance)        
      else
        ! for 2 delta scatterers
        if( SECONDDELTA) call placeSecondDelta(deltaLat,deltaLon,deltaSecondLat, &
                                deltaSecondLon,deltaSecondVertex,deltaSecondDistance)
      endif
                        
      end
       
!-----------------------------------------------------------------------
      subroutine constructPhaseVelocitySquare()
!-----------------------------------------------------------------------   
! determines the phase velocities for all vertices and stores them in an array
! with corresponding indices
!
! returns: phaseVelocitySquare array
      use propagationStartup; use phaseVelocityMap; use parallel
      implicit none
      integer:: i
      real(WP):: getPhaseSquare,lat,lon

      ! store to file              
      if( MASTER ) then 
        open(10,file=datadirectory(1:len_trim(datadirectory))//'PhaseMap.dat')      
        open(11,file=datadirectory(1:len_trim(datadirectory))//'PhaseSquareVertices.dat')
      endif  
        
      ! for all grid vertices
      do i=1,numVertices                                        
        ! determine phase square
        phaseVelocitySquare(i)=getPhaseSquare(i)     
        
        ! file output                  
        if( MASTER ) then 
          call getSphericalCoord_Lat(i,lat,lon)        
          write(10,*) real(lon),real(lat),sqrt(phaseVelocitySquare(i))          
          write(11,*) i,real(phaseVelocitySquare(i))          
        endif        
      enddo
      
      ! close files
      if( MASTER ) then
        close(10)
        close(11)
      endif
      
      end
      
      
!-----------------------------------------------------------------------
      function getPhaseSquare(n)
!-----------------------------------------------------------------------
! calculates the square of phase velocity depending on
! homogeneous, heterogeneous or delta phase maps
! for a certain vertex 
!
! input:
!       n             - vertex index
!
! returns: phase velocity square
      use propagationStartup; use phaseVelocityMap; use parallel; use deltaSecondLocation
      use cells; use verbosity
      implicit none
      integer:: timestep,n
      real(WP):: vectorV(3),vectorS(3),vectorSnd(3),referencePhaseVelocity
      real(WP):: getPhaseSquare,distance,cphase,distanceSnd,lat,lon
            
      ! set reference velocity        
      referencePhaseVelocity=cphaseRef
            
      ! homogeneous phase map (or delta phase map)
      getPhaseSquare = cphaseRef*cphaseRef
 
      !heterogeneous phase velocities
      if( HETEROGENEOUS ) then         
        cphase = phaseMap(n)
        getPhaseSquare = cphase*cphase
        
        !set as new reference for possible delta scatterers that will follow
        referencePhaseVelocity=cphase
      endif
      
      !delta phase map
      if( DELTA ) then
        ! first and second scatterer vectors
        vectorV(:) = vertices(n,:)
        vectorS(:) = vertices(deltaVertex,:)
        call greatCircleDistance(vectorS,vectorV,distance)
        distance=EARTHRADIUS*distance
        if(SECONDDELTA) then
          vectorSnd(:)=vertices(deltaSecondVertex,:)
          call greatCircleDistance(vectorSnd,vectorV,distanceSnd)
          distanceSnd=EARTHRADIUS*distanceSnd
        endif
        
        ! determine which phase velocity to apply for first scatterer
        if( distance .gt. DELTARADIUS) then
          cphase = referencePhaseVelocity
        else
          if(DELTAFUNCTION .eq. "plateau") then
            cphase = referencePhaseVelocity + deltaPerturbation
          else
            if(DELTAFUNCTION .eq. "gaussian") then
              ! some experimental function which looks like a gaussian for distances between 0 and 560 km
              cphase = (1.0-exp(-distance*distance/62500.0))*referencePhaseVelocity+exp(-distance*distance/62500.0)*(referencePhaseVelocity + deltaPerturbation)
            endif
          endif
          
          !write vertex output
          !if(timestep .eq. firsttimestep .and. VERBOSE) then
          !      write(100,*)'distance:',n,distance,DELTARADIUS,cphase
          if( MASTER .and. VERBOSE) then
                print*,'distance scatterer:',n,distance
                call getSphericalCoord_Lat(n,lat,lon)
                print*,'     delta location (lat/lon):',lat,lon
                print*,'     delta radius:',DELTARADIUS,'cphase:',cphase                
          !    !endif
          endif
        endif !distance
        
        ! determine which phase velocity to apply for second scatterer    
        if( SECONDDELTA ) then
          if( distanceSnd .le. DELTARADIUS) then
            if(DELTAFUNCTION .eq. "plateau") then
              cphase = referencePhaseVelocity + deltaPerturbation
            else
              if(DELTAFUNCTION .eq. "gaussian") then
                ! some experimental function which looks like a gaussian for distances between 0 and 560 km
                cphase = (1.0-exp(-distanceSnd*distanceSnd/62500.0))*referencePhaseVelocity+exp(-distanceSnd*distanceSnd/62500.0)*(referencePhaseVelocity + deltaPerturbation)
              endif
            endif
            !write vertex output
            !if(timestep .eq. firsttimestep .and. VERBOSE) then
            !    !  write(100,*)'distance second scatterer:',n,distanceSnd,DELTARADIUS,cphase
            if( MASTER .and. VERBOSE) then
                  print*,'distance second scatterer:',n,distanceSnd
                  call getSphericalCoord_Lat(n,lat,lon)
                  print*,'     delta location (lat/lon):',lat,lon
                  print*,'     delta radius:',DELTARADIUS,'cphase:',cphase
                  print*
            !    !endif
            endif
          endif   
        endif !SECONDDELTA          

        ! squared value for phase velocity
        getPhaseSquare=cphase*cphase            
      endif !DELTA
      
      return
      end

!-----------------------------------------------------------------------
      subroutine printSeismogram()
!-----------------------------------------------------------------------
! plots all seismograms into corresponding files
! file named like:
!     'seismo.*L*.withoutDelta.*#*.dat'    when no delta scatterer was present, else 
!     'seismo.*L*.*lat*.*lon*.dat'
! *L* is wave type, *#* is receiver id, *lat* is latitude, *lon* is longitude of delta scatterer
      use parallel;use griddomain;use propagationStartup; use verbosity
      implicit none
      integer:: i,n,domain,getDomain,strlength,length,ierror
      character*6:: latstr,lonstr
      character*3:: recstr
      character*128:: filename
      real(WP):: reclat,reclon
      external:: getDomain

      write(latstr,'(f6.1)') deltaLat
      write(lonstr,'(f6.1)') deltaLon
      latstr=trim(latstr)
      lonstr=trim(lonstr)
      cphasetype=trim(cphasetype)
      strlength=len_trim(datadirectory)
      
      if(DELTA) then
        filename=datadirectory(1:strlength)//'seismo.'//cphasetype(1:4)//'.'//latstr//'.'//lonstr//'.dat'
      else
        filename=datadirectory(1:strlength)//'seismo.'//cphasetype(1:4)//'.withoutDelta.dat'
      endif      
      
      ! just to remove unneccessary spaces
      filename=trim(filename)        

      if(PARALLELSEISMO) then
        ! each process prints its own output to the current directory
        ! displacement at receiver
        if( manyReceivers ) then
          print*,'    printing to receiver files: '//filename(1:len_trim(filename)-4)
        
          do i=1,size(receivers)
            ! get right seismogram
            domain=getDomain(receivers(i))
            if( rank .eq. domain ) then            
              write(recstr,'(i3.2)') int(i)
              ! get actual vertex position
              call getSphericalCoord_Lat(receivers(i),reclat,reclon)
              
              ! create filename
              if(DELTA) then
                filename=datadirectory(1:strlength)//'seismo.'//cphasetype(1:4)//'.'//latstr//'.'//lonstr//'.'//recstr//'.dat'
              else
                filename=datadirectory(1:strlength)//'seismo.'//cphasetype(1:4)//'.withoutDelta.'//recstr//'.dat'
              endif      
              ! just to remove unneccessary spaces
              filename=trim(filename)        
              
              ! write to file
              open(100+i,file=filename,iostat=ierror)
              if( ierror .ne. 0) call stopProgram('could not open '//filename//'   ')
              write(100+i,*) 'receiver:',i,real(reclat),real(reclon),receivers(i)
              do n=1,size( receiversSeismogram(size(receivers)+1,:) )
                write(100+i,*) receiversSeismogram(size(receivers)+1,n),receiversSeismogram(i,n)
              enddo
              close(100+i)
            endif
          enddo
        else
          ! get right seismogram
          domain=getDomain(receiverVertex)
          if( domain .ne. 0) then
            if( MASTER ) then          
              ! receive from domain process
              length=size(seismogramReceiver(:,:))
              tag = domain
              !debug
              if(DEBUG) print*,'getting seismogram from domain:',domain,length              
              call MPI_Recv(seismogramReceiver,length,MPI_CUSTOM,domain,tag,MPI_COMM_WORLD,status,ierror)
            else if( rank .eq. domain) then
              ! send seismogram to master
              length=size(seismogramReceiver(:,:))
              tag = rank
              if(DEBUG) print*,rank,'sending receiver seismogram...'
              call MPI_Send(seismogramReceiver,length,MPI_CUSTOM,0,tag,MPI_COMM_WORLD,ierror)
            endif            
          endif
        
          ! output
          if( MASTER ) then
            ! displacement at receiver
            print*,'    printing to file: '//filename          
            open(100,file=filename)
            do n=1,size(seismogramReceiver(1,:))
              write(100,*) seismogramReceiver(1,n),seismogramReceiver(2,n)
            enddo
            close(100)
          endif
        endif
      else
        ! each process prints its own output to the current directory
        ! displacement at receiver
        if( manyReceivers ) then
          print*,'    printing to receiver files: '//filename(1:len_trim(filename)-4)
        
          do i=1,size(receivers)
            write(recstr,'(i3.2)') int(i)
            ! get actual vertex position
            call getSphericalCoord_Lat(receivers(i),reclat,reclon)
            
            ! create filename
            if(DELTA) then
              filename=datadirectory(1:strlength)//'seismo.'//cphasetype(1:4)//'.'//latstr//'.'//lonstr//'.'//recstr//'.dat'
            else
              filename=datadirectory(1:strlength)//'seismo.'//cphasetype(1:4)//'.withoutDelta.'//recstr//'.dat'
            endif      
            ! just to remove unneccessary spaces
            filename=trim(filename)        
            
            ! write to file
            open(100+i,file=filename,iostat=ierror)
            if( ierror .ne. 0) call stopProgram('could not open '//filename//'   ')
            write(100+i,*) 'receiver:',i,real(reclat),real(reclon),receivers(i)
            do n=1,size( receiversSeismogram(size(receivers)+1,:) )
              write(100+i,*) receiversSeismogram(size(receivers)+1,n),receiversSeismogram(i,n)
            enddo
            close(100+i)        
          enddo
        else
          print*,'    printing to file: '//filename
          open(100,file=filename)
          do n=1,size(seismogramReceiver(1,:))
            write(100,*) seismogramReceiver(1,n),seismogramReceiver(2,n)
          enddo
          close(100)
        endif
      endif
      end
